'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var extension = require('../../extension');
var view = require('../../view');
var rangeset = require('../../rangeset');
var state = require('../../state');
var styleMod = require('style-mod');

/// A gutter marker represents a bit of information attached to a line
/// in a specific gutter. Your own custom markers have to extend this
/// class.
class GutterMarker extends rangeset.RangeValue {
    /// @internal
    compare(other) {
        return this == other || this.constructor == other.constructor && this.eq(other);
    }
    /// Map this marker through a position mapping.
    map(mapping, pos) {
        pos = mapping.mapPos(pos, -1, state.MapMode.TrackBefore);
        return pos < 0 ? null : new rangeset.Range(pos, pos, this);
    }
    /// Render the DOM node for this marker, if any.
    toDOM() { return null; }
    /// Create a range that places this marker at the given position.
    at(pos) { return new rangeset.Range(pos, pos, this); }
}
GutterMarker.prototype.elementClass = "";
const defaults = {
    fixed: true,
    renderEmptyElements: false,
    elementClass: "",
    initialMarkers: () => rangeset.RangeSet.empty,
    updateMarkers: (markers) => markers,
    lineMarker: () => null,
    initialSpacer: null,
    updateSpacer: null
};
/// Create a gutter extension.
function gutter(config) {
    let conf = extension.fillConfig(config, defaults);
    // FIXME allow client code to preserve a gutter config
    let plugin = new view.ViewPlugin(view => new GutterView(view, conf));
    return [plugin.extension, view.EditorView.styleModule(styles)];
}
class GutterView {
    constructor(view, config) {
        this.view = view;
        this.config = config;
        this.elements = [];
        this.spacer = null;
        this.dom = document.createElement("div");
        this.dom.className = "codemirror-gutter " + config.class + " " + styles.gutter;
        this.dom.setAttribute("aria-hidden", "true");
        if (config.fixed) {
            // FIXME IE11 fallback, which doesn't support position: sticky,
            // by using position: relative + event handlers that realign the
            // gutter (or just force fixed=false on IE11?)
            this.dom.style.position = "sticky";
        }
        view.dom.insertBefore(this.dom, view.contentDOM);
        this.markers = config.initialMarkers(view);
        if (config.initialSpacer) {
            this.spacer = new GutterElement(0, 0, [config.initialSpacer(view)], this.config.elementClass);
            this.dom.appendChild(this.spacer.dom);
            this.spacer.dom.style.cssText += "visibility: hidden; pointer-events: none";
        }
    }
    update(update) {
        this.markers = this.config.updateMarkers(this.markers.map(update.changes), update);
        if (this.spacer && this.config.updateSpacer) {
            let updated = this.config.updateSpacer(this.spacer.markers[0], update);
            if (updated != this.spacer.markers[0])
                this.spacer.update(0, 0, [updated], this.config.elementClass);
        }
    }
    draw() {
        // FIXME would be nice to be able to recognize updates that didn't redraw
        let i = 0, height = 0;
        let markers = this.markers.iter(this.view.viewport.from, this.view.viewport.to);
        let localMarkers = [], nextMarker = markers.next();
        this.view.viewportLines(line => {
            let text;
            if (Array.isArray(line.type))
                text = line.type.find(b => b.type == view.BlockType.Text);
            else
                text = line.type == view.BlockType.Text ? line : undefined;
            if (!text)
                return;
            while (nextMarker && nextMarker.from <= text.from) {
                if (nextMarker.from == text.from)
                    localMarkers.push(nextMarker.value);
                nextMarker = markers.next();
            }
            let forLine = this.config.lineMarker(this.view, text, localMarkers);
            if (forLine)
                localMarkers.unshift(forLine);
            if (localMarkers.length || this.config.renderEmptyElements) {
                let above = text.top - height;
                if (i == this.elements.length) {
                    let newElt = new GutterElement(text.height, above, localMarkers, this.config.elementClass);
                    this.elements.push(newElt);
                    this.dom.appendChild(newElt.dom);
                }
                else {
                    let markers = localMarkers, elt = this.elements[i];
                    if (sameMarkers(markers, elt.markers)) {
                        markers = elt.markers;
                        localMarkers.length = 0;
                    }
                    elt.update(text.height, above, markers, this.config.elementClass);
                }
                height = text.bottom;
                i++;
                if (localMarkers.length)
                    localMarkers = [];
            }
        }, 0);
        while (this.elements.length > i)
            this.dom.removeChild(this.elements.pop().dom);
        this.dom.style.minHeight = this.view.contentHeight + "px";
    }
    destroy() {
        this.dom.remove();
    }
    get styles() { return styles; }
}
class GutterElement {
    constructor(height, above, markers, eltClass) {
        this.height = -1;
        this.above = 0;
        this.dom = document.createElement("div");
        this.update(height, above, markers, eltClass);
    }
    update(height, above, markers, eltClass) {
        if (this.height != height)
            this.dom.style.height = (this.height = height) + "px";
        if (this.above != above)
            this.dom.style.marginTop = (this.above = above) ? above + "px" : "";
        if (this.markers != markers) {
            this.markers = markers;
            for (let ch; ch = this.dom.lastChild;)
                ch.remove();
            let cls = "codemirror-gutter-element " + styles.gutterElement;
            if (eltClass)
                cls += " " + eltClass;
            for (let m of markers) {
                let dom = m.toDOM();
                if (dom)
                    this.dom.appendChild(dom);
                let c = m.elementClass;
                if (c)
                    cls += " " + c;
            }
            this.dom.className = cls;
        }
    }
}
function sameMarkers(a, b) {
    if (a.length != b.length)
        return false;
    for (let i = 0; i < a.length; i++)
        if (!a[i].compare(b[i]))
            return false;
    return true;
}
/// Used to insert markers into the line number gutter.
const lineNumberMarkers = extension.Slot.define();
/// Create a line number gutter extension. The order in which the
/// gutters appear is determined by their extension priority.
const lineNumbers = view.EditorView.extend.unique(configs => {
    let config = extension.combineConfig(configs, {
        fixed: true,
        formatNumber: String
    });
    class NumberMarker extends GutterMarker {
        constructor(number) {
            super();
            this.number = number;
        }
        eq(other) { return this.number == other.number; }
        toDOM() {
            return document.createTextNode(config.formatNumber(this.number));
        }
    }
    return gutter({
        class: "codemirror-line-numbers " + styles.lineNumberGutter,
        fixed: config.fixed,
        elementClass: styles.lineNumberGutterElement,
        updateMarkers(markers, update) {
            let slot = update.getMeta(lineNumberMarkers);
            if (slot)
                markers = markers.update(slot.add || [], slot.filter || null);
            return markers;
        },
        lineMarker(view, line, others) {
            if (others.length)
                return null;
            // FIXME try to make the line number queries cheaper?
            return new NumberMarker(view.state.doc.lineAt(line.from).number);
        },
        initialSpacer(view) {
            return new NumberMarker(maxLineNumber(view.state.doc.lines));
        },
        updateSpacer(spacer, update) {
            let max = maxLineNumber(update.view.state.doc.lines);
            return max == spacer.number ? spacer : new NumberMarker(max);
        }
    });
}, {});
function maxLineNumber(lines) {
    let last = 9;
    while (last < lines)
        last = last * 10 + 9;
    return last;
}
const styles = new styleMod.StyleModule({
    gutter: {
        display: "flex !important",
        flexDirection: "column",
        flexShrink: 0,
        left: 0,
        boxSizing: "border-box",
        height: "100%",
        overflow: "hidden"
    },
    gutterElement: {
        boxSizing: "border-box"
    },
    lineNumberGutter: {
        background: "#f5f5f5",
        borderRight: "1px solid silver"
    },
    lineNumberGutterElement: {
        padding: "0 3px 0 5px",
        minWidth: "20px",
        textAlign: "right",
        color: "#999",
        whiteSpace: "nowrap"
    }
});

exports.GutterMarker = GutterMarker;
exports.gutter = gutter;
exports.lineNumberMarkers = lineNumberMarkers;
exports.lineNumbers = lineNumbers;
